__name__ = "biotite2pymol"
__author__ = "Patrick Kunzmann"
__all__ = ["launch_pymol", "setup_parameters"]

import pymol
from pymol import cmd as default_cmd


def launch_pymol(*args):
    """
    Launch *PyMOL* with the given command line arguments.

    This is the recommended way to start *PyMOL* in GUI mode.
    It starts *PyMOL* by calling :func:`pymol.finish_launching()`,
    reinitializes *PyMOL* to clear the workspace and sets up necessary
    parameters using :func:`setup_parameters()`.

    Parameters
    ----------
    *args : str
        The command line options given to *PyMOL*.
    """
    pymol.finish_launching(["pymol"] + list(args))
    default_cmd.reinitialize()
    setup_parameters()


def setup_parameters(pymol_instance=None):
    """
    Sets *PyMOL* parameters that are necessary for *biotite2pymol* to interact
    properly with *PyMOL*.

    pymol_instance : PyMOL, optional
        When using the object-oriented *PyMOL* API the :class:`PyMOL`
        object must be given here.
    """
    if pymol_instance is None:
        cmd = default_cmd
    else:
        cmd = pymol_instance.cmd
    
    # The selections only work properly,
    # if the order stays the same after adding a model to PyMOL
    cmd.set("retain_order", 1)


__name__ = "biotite2pymol"
__author__ = "Patrick Kunzmann"
__all__ = ["to_biotite", "to_pymol",
           "convert_to_atom_array", "convert_to_chempy_model"]

import warnings
import numpy as np
import biotite
from biotite.sequence import ProteinSequence
import biotite.structure as struc
import biotite.structure.io as strucio
import pymol
from pymol import cmd as default_cmd
from chempy.models import Indexed as IndexedModel
from chempy import Atom, Bond


def to_biotite(object_name, state=None, extra_fields=None,
               include_bonds=False, pymol_instance=None):
    """
    Convert a *PyMOL* object into an :class:`AtomArray` or
    :class:`AtomArrayStack`.

    Parameters
    ----------
    object_name : str
        The name of the *PyMOL* object to be converted.
    state : int, optional
        If this parameter is given, the function will return an
        :class:`AtomArray` corresponding to the given state of the
        *PyMOL* object.
        If this parameter is omitted, an :class:`AtomArrayStack`
        containing all states will be returned, even if the *PyMOL*
        object contains only one state.
    extra_fields : list of str, optional
        The strings in the list are optional annotation categories
        that should be stored in the output atom array (stack).
        ``'b_factor'``, ``'occupancy'`` and``'charge'`` are valid
        values.
    include_bonds : bool, optional
        If set to true, an associated :class:`BondList` will be created
        for the returned atom array (stack).
    pymol_instance : PyMOL, optional
        When using the object-oriented *PyMOL* API the :class:`PyMOL`
        object must be given here.
    
    Returns
    -------
    structure : AtomArray or AtomArrayStack
        The converted structure.
        Wheather an :class:`AtomArray` or :class:`AtomArrayStack` is
        returned depends on the `state` parameter.
    """
    if pymol_instance is None:
        cmd = default_cmd
    else:
        cmd = pymol_instance.cmd
    
    if state is None:
        model = cmd.get_model(object_name, state=1)
        template = convert_to_atom_array(model, extra_fields, include_bonds)
        coord = np.stack(
            [cmd.get_coordset(object_name, state=i+1)
             for i in range(cmd.count_states(object_name))]
        )
        return struc.from_template(template, coord)
    else:
        model = cmd.get_model(object_name, state=state)
        return convert_to_atom_array(model, extra_fields, include_bonds)


def to_pymol(object_name, atoms, pymol_instance=None):
    """
    Convert an :class:`AtomArray` or :class:`AtomArrayStack` into a
    *PyMOL* object and add it to the *PyMOL* session.

    Parameters
    ----------
    object_name : str
        The name of the newly created *PyMOL* object.
    atoms : AtomArray or AtomArrayStack
        The structure to be converted.
    pymol_instance : PyMOL, optional
        When using the object-oriented *PyMOL* API the :class:`PyMOL`
        object must be given here.
    """
    if pymol_instance is None:
        cmd = default_cmd
    else:
        cmd = pymol_instance.cmd
    
    if isinstance(atoms, struc.AtomArray) or \
       (isinstance(atoms, struc.AtomArrayStack) and atoms.stack_depth == 1):
            model = convert_to_chempy_model(atoms)
            cmd.load_model(model, object_name)
    elif isinstance(atoms, struc.AtomArrayStack):
        # Use first model as template
        model = convert_to_chempy_model(atoms[0])
        cmd.load_model(model, object_name)
        # Append states corresponding to all following models
        for coord in atoms.coord[1:]:
            cmd.load_coordset(coord, object_name)
    else:
        raise TypeError("Expected 'AtomArray' or 'AtomArrayStack'")


def convert_to_atom_array(chempy_model, extra_fields=None, 
                          include_bonds=False):
    """
    Convert a :class:`chempy.models.Indexed`
    object into an :class:`AtomArray`.

    Parameters
    ----------
    chempy_model : Indexed
        The ``chempy`` model.
    extra_fields : list of str, optional
        The strings in the list are optional annotation categories
        that should be stored in the output atom array.
        ``'b_factor'``, ``'occupancy'`` and``'charge'`` are valid
        values.
    include_bonds : bool, optional
        If set to true, an associated :class:`BondList` will be created
        for the returned atom array.
    
    Returns
    -------
    atom_array : AtomArray
        The converted structure.
    """
    if extra_fields is None:
        extra_fields = []
    
    
    atoms = chempy_model.atom
    # Ignore alternative locations
    atoms = [atom for atom in atoms if atom.alt in ("", "A")]

    bonds = chempy_model.bond

    
    atom_array = struc.AtomArray(len(atoms))
    
    
    atom_array.chain_id = np.array(
        [a.chain for a in atoms],
        dtype="U3"
    )
    atom_array.res_id = np.array(
        [a.resi_number for a in atoms],
        dtype=int
    )
    atom_array.ins_code = np.array(
        [a.ins_code for a in atoms],
        dtype="U1"
    )
    atom_array.res_name = np.array(
        [a.resn for a in atoms],
        dtype="U3"
    )
    atom_array.hetero = np.array(
        [a.hetatm for a in atoms],
        dtype=bool
    )
    atom_array.atom_name = np.array(
        [a.name for a in atoms],
        dtype="U6"
    )
    atom_array.element = np.array(
        [a.symbol for a in atoms],
        dtype="U2"
    )
    
    if "b_factor" in extra_fields:
        atom_array.b_factor = np.array(
            [a.b for a in atoms],
            dtype=float
        )
    if "occupancy" in extra_fields:
        atom_array.occupancy = np.array(
            [a.q for a in atoms],
            dtype=float
        )
    if "charge" in extra_fields:
        atom_array.charge = np.array(
            [a.formal_charge for a in atoms],
            dtype=int
        )
    

    atom_array.coord = np.array(
        [a.coord for a in atoms],
        dtype=np.float32
    )


    if include_bonds:
        bond_array = np.array(
            [[b.index[0], b.index[1], b.order] for b in bonds],
            dtype=np.uint32
        )
        atom_array.bonds = struc.BondList(len(atoms), bond_array)
    

    return atom_array


def convert_to_chempy_model(atom_array):
    """
    Convert an :class:`AtomArray` into a :class:`chempy.models.Indexed`
    object.

    Returns
    -------
    chempy_model : Indexed
        The converted structure.
    """
    model = IndexedModel()

    annot_cat = atom_array.get_annotation_categories()


    for i in range(atom_array.array_length()):
        atom = Atom()

        atom.segi = atom_array.chain_id[i]
        atom.chain = atom_array.chain_id[i]

        atom.resi_number = atom_array.res_id[i]

        atom.ins_code = atom_array.ins_code[i]
        
        res_name = atom_array.res_name[i]
        atom.resn = res_name
        if len(res_name) == 1:
            atom.resn_code = res_name
        else:
            try:
                atom.resn_code = ProteinSequence.convert_letter_3to1(res_name)
            except KeyError:
                atom.resn_code = "X"

        atom.hetatm = 1 if atom_array.hetero[i] else 0

        atom.name = atom_array.atom_name[i]

        atom.symbol = atom_array.element[i]
        
        if "b_factor" in annot_cat:
            atom.b = atom_array.b_factor[i]
        
        if "occupancy" in annot_cat:
            atom.q = atom_array.occupancy[i]
        
        if "charge" in annot_cat:
            atom.formal_charge = atom_array.charge[i]
        
        atom.coord = tuple(atom_array.coord[..., i, :])

        atom.index = i+1

        model.add_atom(atom)
    

    if atom_array.bonds is not None:
        for i, j, order in atom_array.bonds.as_array():
            bond = Bond()

            if order != 0:
                bond.order = order
            else:
                # If bond order is not defined,
                # use a single bond as default
                bond.order = 1
            
            bond.index = [i, j]

            model.add_bond(bond)
    else:
        warnings.warn(
            "The given atom array (stack) has no associated bond information"
        )
    

    return model


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Transfer structure data from Biotite to PyMOL.
"""

__version__ = "0.1.1"
__name__ = "biotite2pymol"
__author__ = "Patrick Kunzmann"

from .display import *
from .model import *
from .selection import *
from .startup import *


__name__ = "biotite2pymol"
__author__ = "Patrick Kunzmann"
__all__ = ["select"]

import numpy as np


def select(object_name, mask):
    """
    Convert a boolean mask for atom selection into a *PyMOL* selection
    string.

    Parameters
    ----------
    object_name : str
        The name of the *PyMOL* object to select the atoms from.
        Usually this is the same name used in a previous
        :func:`to_pymol()` call, where the atom array (stack) the given
        `mask` corresponds to was transferred to *PyMOL*.
    mask : ndarray, dtype=bool
        The boolean mask to be converted into a selection string.
    
    Returns
    -------
    selection : str
        A *PyMOL* compatible selection string.
    """
    # Indices where the mask changes from True to False
    # or from False to True
    # The '+1' makes each index refer to the position after the change
    # i.e. the new value
    changes = np.where(np.diff(mask))[0] + 1
    # If first element is True, insert index 0 at start
    # -> the first change is always from False to True
    if mask[0]:
        changes = np.concatenate(([0], changes))
    # If the last element is True, insert append length of mask
    # as exclusive stop index
    # -> the last change is always from True to False
    if mask[-1]:
        changes = np.concatenate((changes, [len(mask)]))
    # -> Changes are alternating (F->T, T->F, F->T, ..., F->T, T->F)
    # Reshape into pairs of changes ([F->T, T->F], [F->T, T->F], ...)
    # -> these are the intervals where the mask is True
    intervals = np.reshape(changes, (-1, 2))

    # Convert interval into selection string
    # Two things to note:
    # - PyMOLs indexing starts at 1-> 'start+1'
    # - Stop index in 'intervals' is exclusive -> 'stop+1-1' -> 'stop'
    index_selection = " or ".join(
        [f"index {start+1}-{stop}" for start, stop in intervals]
    )
    # Constrain the selection to given object name
    complete_selection = f"model {object_name} and ({index_selection})"
    return complete_selection


__name__ = "biotite2pymol"
__author__ = "Patrick Kunzmann"
__all__ = ["show", "TimeoutError"]

import tempfile
import time
import datetime
from os.path import join, getsize
import pymol
from pymol import cmd as default_cmd


INTERVAL = 0.1


def show(size=None, use_ray=False, timeout=60.0, pymol_instance=None):
    """
    Render an image of the PyMOL session and display it in the current
    *Jupyter* notebook.

    Note that this function works only in a *Jupyter* notebook.

    Parameters
    ----------
    size : tuple of (int, int), optional
        The width and height of the rendered image.
        By default, the size of the current *PyMOL* viewport is used
    use_ray : bool, optional
        If set to true, the a ray-traced image is created.
        This will also increase the rendering time.
    timeout : float
        The number of seconds to wait for image output from *PyMOL*. 
    pymol_instance : PyMOL, optional
        When using the object-oriented *PyMOL* API the :class:`PyMOL`
        object must be given here.
    
    Raises
    ------
    TimeoutError
        If no image was created after expiry of the `timeout` limit.
    """
    try:
        from IPython.display import Image
    except ImportError:
        raise ImportError("IPython is not installed")

    if pymol_instance is None:
        cmd = default_cmd
    else:
        cmd = pymol_instance.cmd
    
    if size is None:
        width = 0
        height = 0
    else:
        width, height = size
    
    if use_ray:
        ray = 1
    else:
        ray = 0
    
    image_file = tempfile.NamedTemporaryFile(
        delete=False, prefix="biotite2pymol", suffix=".png"
    )
    image_file.close()
    
    start_time = datetime.datetime.now()

    cmd.png(image_file.name, width, height, ray=ray)
    
    while True:
        # After 'timeout' seconds the loop exits with an error
        if (datetime.datetime.now() - start_time).total_seconds() > timeout:
            raise TimeoutError(
                "No PNG image was output within the expected time limit"
            )
        
        # Check if PyMOL has already written image data to file
        if getsize(image_file.name) > 0:
            break

        time.sleep(INTERVAL)
    
    return Image(image_file.name)


def TimeoutError(Exception):
    """
    Exception that is raised after time limit expiry in :func:`show()`.
    """
    pass


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['biotite2pymol']

package_data = \
{'': ['*']}

install_requires = \
['biotite>=0.21,<0.22']

setup_kwargs = {
    'name': 'ammolite',
    'version': '0.1.1',
    'description': 'Visualize structure data from Biotite with PyMOL',
    'long_description': ".. image:: https://raw.githubusercontent.com/biotite-dev/biotite2pymol/master/doc/static/assets/biotite2pymol_logo_s.png\n  :alt: biotite2pymol logo\n  :align: center\n\nbiotite2pymol - From Biotite to PyMOL and back again\n====================================================\n\nThis package enables the transfer of structure related objects\nfrom `Biotite <https://www.biotite-python.org/>`_\nto `PyMOL <https://pymol.org/>`_ for visualization,\nvia PyMOL's Python API:\n\n- Import ``AtomArray`` and ``AtomArrayStack`` objects into *PyMOL* -\n  without intermediate structure files.\n- Convert *PyMOL* objects into ``AtomArray`` and ``AtomArrayStack`` instances.\n- Use *Biotite*'s boolean masks for atom selection in *PyMOL*.\n\nHave a look at the `example Jupyter notebook <https://github.com/biotite-dev/biotite2pymol/blob/master/doc/examples/cytochrome.ipynb>`_.\n\n|\n\n.. image:: https://raw.githubusercontent.com/biotite-dev/biotite2pymol/master/doc/demo/demo.gif\n    :alt: biotite2pymol demo\n\n|\n\n*PyMOL is a trademark of Schrodinger, LLC.*",
    'author': 'Patrick Kunzmann',
    'author_email': 'patrick.kunzm@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://ammolite.biotite-python.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
